import logging
from typing import List

import pkg_resources
import yaml
from click import command, option, Choice
from jsonasobj2 import as_dict
from linkml_runtime import SchemaView

from schemasheets.conf.configschema import ColumnSettings
from schemasheets.schema_exporter import SchemaExporter
from schemasheets.schemasheet_datamodel import TableConfig, ColumnConfig
from typing import List, Dict

# todo write tests

# todo: include read only slots?

# todo: check metamodel if element is deprecated
#  (if it can be included in the report, the deprecation status will be reported!)
#   like subclass_of, subproperty_of ?

# todo check for inlining in addition to checking if a range class has a identifier slot?

# todo support generating an exhaustive and concise report in the same run

# todo: support slots which might require multiple columns for multiple inner keys,
#  like examples (values and description)

root_classes = ['slot_definition', "class_definition"]  # hard coding the intention to do a class slot usage report

# todo could these be semantically derived from the root_classes? other than lexical string splitting?
boilerplate_cols = ['slot', 'class']

# these a slots whose ranges are classes with identifier slots, but they still can't be included in the report
blacklist = [
    'attributes',
    'instantiates',
    'name',
    'slot_usage',
    'slots',
]

# this salvages these slots (by name) from untemplateables
requires_column_settings = {
    "examples values": {
        "name": "examples",
        "ikm_slot": "value",  # todo: what about description?
        "ikm_class": "example",
        "internal_separator": "|",

    },
    "structured_pattern": {
        "name": "structured_pattern",
        "ikm_slot": "syntax",  # todo what about interpolated and partial_match
        "ikm_class": "pattern_expression",
        "internal_separator": "|",

    },
    "unit symbol": {
        "name": "unit",
        "ikm_slot": "symbol",
        # todo what about symbol, exact mappings, ucum_code, derivation, has_quantity_kind, iec61360code
        "ikm_class": "UnitOfMeasure",
        "internal_separator": "|",
    },
}

debug_report = {
    "blacklist_skipped": []
}


def setup_logging(log_file=None, log_level=logging.INFO):
    logging.basicConfig(
        level=log_level,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        filename=log_file,
    )


def tabulate_unique_values(value_list):
    """
    tabulate the number of appearances of each unique values in a list

    :param value_list: The list to tabulate, potentially with repeat occurrences of list items.
    :returns: A dict mapping each unique value to the number of times it appears in the list.
    """

    unique_values = set(value_list)
    value_counts = {}
    for value in unique_values:
        count = value_list.count(value)
        value_counts[value] = count

    sorted_value_counts = sorted(value_counts.items(), key=lambda x: x[1], reverse=True)

    return sorted_value_counts


def discover_source_usage(source_view: SchemaView) -> tuple[List[str], List[str]]:
    """
    Discover the meta slots and annotations used in the source code.

    :param source_view: A SchemaView of the source schema.
    :returns: A tuple of the meta slots and annotations that were discovered as used in the source schema.
    """

    discovered_meta_slots = []
    discovered_annotations = []
    source_classes = source_view.all_classes()
    for ck, cv in source_classes.items():
        cv_dict = cv.__dict__
        for cvdk, cvdv in cv_dict.items():
            if cvdv:
                discovered_meta_slots.append(cvdk)

        class_annotations = list(source_view.get_class(ck).annotations.keys())  # TODO this isn't well tested yet
        for ca in class_annotations:
            discovered_annotations.append(ca)

        for cis in source_view.class_induced_slots(ck):
            cis_dict = cis.__dict__

            for cisk, cisv in cis_dict.items():
                if cisv:
                    discovered_meta_slots.append(cisk)

            cis_annotations_dict = as_dict(cis.annotations)
            for a in cis_annotations_dict:
                discovered_annotations.append(a)

    return discovered_meta_slots, discovered_annotations


def do_usage_report(
        style: str,
        meta_view: SchemaView,
        meta_type_names: List[str],
        meta_enum_names: List[str],
        discovered_annotations: List[str],
        discovered_source_slots: List[str],
        logger: logging.Logger,
) -> tuple[TableConfig, Dict[str, Dict[str, str]]]:
    """
    Perform a usage report on the metamodel.

    :param style: The style of the usage report.
    :param meta_view: The metamodel to analyze.
    :param meta_type_names: The list of types discovered in the metamodel.
    :param meta_enum_names: The list of enums discovered in the metamodel.
    :param discovered_annotations: The list of annotations discovered in the source schema.
    :param discovered_source_slots: The list of metaslots used in the source 's SLotDefinitions and ClassDefinitions.
    :param logger: The logger to use.
    :returns: A tuple of the table config and the untemplateable metaslots.
    """

    columns_dict = {}

    # we discover which class-range slots have an identifying slot and then exclude the blacklist slots
    identifiables = {}

    slot_scan_results = {}

    # these are slots whose range is a class lacking an identifying slot
    untemplateables = {}

    discovered_cols = []
    slot_ranges = []

    discovered_annotations = list(set(discovered_annotations))

    discovered_source_slots = list(set(discovered_source_slots) - set(blacklist))
    discovered_source_slots = list(set(discovered_source_slots))
    discovered_source_slots.sort()
    discovered_annotations.sort()
    discovered_source_slots.sort()

    for current_root in root_classes:
        current_induced_slots = meta_view.class_induced_slots(current_root)
        for cis in current_induced_slots:

            temp_dict = {
                "range": cis.range,
                "multivalued": cis.multivalued,
                "type_range": cis.range in meta_type_names,
            }
            if style == "exhaustive" or cis.name in discovered_source_slots:
                if cis.name not in slot_scan_results:
                    slot_scan_results[cis.name] = temp_dict

                else:
                    if slot_scan_results[cis.name] == temp_dict:
                        continue
                    else:
                        logger.warning(f"Redefining {cis.name} from {slot_scan_results[cis.name]} to {temp_dict}")

    for c in boilerplate_cols:
        columns_dict[c] = ColumnConfig(name=c,
                                       maps_to=c,
                                       is_element_type=True,
                                       settings=ColumnSettings()
                                       )

    for rcs_k, rcs_v in requires_column_settings.items():
        if style == "exhaustive" or rcs_v['name'] in discovered_source_slots:
            temp_settings = ColumnSettings()
            if "internal_separator" in rcs_v:
                temp_settings.internal_separator = rcs_v["internal_separator"]
            if "ikm_class" in rcs_v and "ikm_slot" in rcs_v:
                temp_settings.inner_key = rcs_v["ikm_slot"]
            columns_dict[rcs_k] = ColumnConfig(
                name=rcs_v["name"],
                is_element_type=False,
                maps_to=rcs_v["name"],
                metaslot=meta_view.get_slot(rcs_v["name"]),
                settings=temp_settings
            )
            if "ikm_class" in rcs_v and "ikm_slot" in rcs_v:
                columns_dict[rcs_k].inner_key_metaslot = meta_view.induced_slot(rcs_v["ikm_slot"], rcs_v["ikm_class"])

    for ssk, ssv in slot_scan_results.items():
        if ssv["range"] not in meta_type_names and ssv["range"] not in meta_enum_names:
            slot_ranges.append(ssv["range"])

    slot_ranges.sort()

    tabulation_results = tabulate_unique_values(slot_ranges)

    for tabulation_result in tabulation_results:
        current_range = tabulation_result[0]
        # current_count = tabulation_result[1]
        current_identifier = meta_view.get_identifier_slot(current_range)
        if current_identifier:
            identifiables[current_range] = current_identifier.name
        else:
            pass

    requires_column_settings_names = []
    for rcs_k, rcs_v in requires_column_settings.items():
        requires_column_settings_names.append(rcs_v["name"])

    silenced = requires_column_settings_names + ["annotations"]

    for ssrk, ssrv in slot_scan_results.items():
        current_range = ssrv["range"]
        if current_range in meta_type_names or current_range in meta_enum_names:
            discovered_cols.append(ssrk)
        elif current_range in identifiables:
            if ssrk in blacklist:
                debug_report['blacklist_skipped'].append(ssrk)
            else:
                discovered_cols.append(ssrk)
        elif ssrk in silenced:
            continue
        else:
            untemplateables[ssrk] = ssrv

    for da in discovered_annotations:
        columns_dict[da] = ColumnConfig(
            name="annotations",
            is_element_type=False,
            maps_to="annotations",
            metaslot=meta_view.get_slot("annotations"),
            settings=ColumnSettings(inner_key=da),
        )

    discovered_cols.sort()

    for c in discovered_cols:
        c_slug = c.replace(" ", "_")
        ms = meta_view.get_slot(c)
        mv = False
        if ms:
            mv = ms.multivalued
        if mv:
            cs = ColumnSettings(internal_separator="|")
        else:
            cs = ColumnSettings()
        cc = ColumnConfig(
            maps_to=c,
            metaslot=ms,
            name=c_slug,
            settings=cs,
        )
        columns_dict[c] = cc  # use verbatim elements (not slugged) in first row?

    new_tc = TableConfig(
        column_by_element_type={'slot': 'slot', 'class': 'class'},
        columns=columns_dict
    )

    return new_tc, untemplateables


@command()
@option("--verbose", is_flag=True, help="Enable verbose logging.")
@option("--source-path", "-i", required=True,
        help="A filesystem or URL path to the schema that should be analysed and reported.")
@option("--output-path", "-o", required=True)
@option("--debug-report-path", "-d")
@option("--log-file", "-l")
@option("--report-style", "-s", type=Choice(['exhaustive', 'concise']), required=True)
def cli(source_path, output_path, debug_report_path, verbose, log_file, report_style):
    """
    A CLI tool to generate a slot usage schemasheet from the LinkML metamodel and a source schema.

    :param source_path: A filesystem or URL path to the schema that should be analysed and reported
    :param output_path: The filesystem path where the generated schemasheet will be written.
    :param debug_report_path: An optional filesystem path where a YAML report of excluded metaslots will be saved.
    :param verbose: If true, DEBUG logging will be used.
    :param log_file: An optional filesystem path where log messages will be saved.
    :param report_style: exhaustive means that all non-excluded metaslots will be included. concise means that only those metaslots used by the source schema will be included.
    """

    # in some cases it will be better to get this from a local filesystem, not a URL...
    # todo: add script/targets for downloading and merging

    if verbose:
        setup_logging(log_file, logging.DEBUG)
    else:
        setup_logging(log_file)

    logger = logging.getLogger(__name__)

    meta_yaml_path = pkg_resources.resource_filename(
        'linkml_runtime',
        'linkml_model/model/schema/meta.yaml'
    )  # todo proper platform agnostic path

    meta_view = SchemaView(meta_yaml_path, merge_imports=True)

    source_view = SchemaView(source_path, merge_imports=True)

    (discovered_source_slots, discovered_annotations) = discover_source_usage(source_view)

    root_classes.sort()

    meta_types = meta_view.schema.types
    meta_type_names = list(meta_types.keys())

    meta_enum_names = list(meta_view.all_enums().keys())
    meta_enum_names.sort()

    new_tc, untemplateables = do_usage_report(
        discovered_annotations=discovered_annotations,
        discovered_source_slots=discovered_source_slots,
        logger=logger,
        meta_enum_names=meta_enum_names,
        meta_type_names=meta_type_names,
        meta_view=meta_view,
        style=report_style,
    )

    current_exporter = SchemaExporter()
    current_exporter.export(
        schemaview=source_view,
        table_config=new_tc,
        to_file=output_path,
    )

    untemplateable_report = {}
    for uk, uv in untemplateables.items():
        untemplateable_report[uk] = uv["range"]

    debug_report['blacklist'] = blacklist
    debug_report['untemplateables'] = untemplateable_report
    debug_report['untemplateable_skipped'] = list(set(untemplateables).intersection(set(discovered_source_slots)))

    if debug_report_path:
        try:
            with open(debug_report_path, 'w') as yaml_file:
                yaml.safe_dump(debug_report, yaml_file)
            logger.warning(f"Successfully dumped the debug_report to '{debug_report_path}'.")
        except Exception as e:
            logger.warning(f"An error occurred: {e}")


if __name__ == "__main__":
    cli()



import csv
import logging
import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, Any, List, Optional, TextIO, Union

import click
from linkml_runtime.linkml_model import Element, SlotDefinition, SubsetDefinition, ClassDefinition, EnumDefinition, \
    PermissibleValue, \
    TypeDefinition, Example, Annotation, Prefix
from linkml_runtime.utils.formatutils import underscore
from linkml_runtime.utils.schemaview import SchemaView

from schemasheets.conf.configschema import ColumnSettings
from schemasheets.schemamaker import SchemaMaker
from schemasheets.schemasheet_datamodel import TableConfig, T_CLASS, T_SLOT, SchemaSheet, T_ENUM, T_PV, T_TYPE, \
    T_SUBSET, T_PREFIX

ROW = Dict[str, Any]


def _configuration_has_primary_keys_for(table_config: TableConfig, metatype: str) -> bool:
    for col_name, col_config in table_config.columns.items():
        if col_config.is_element_type and col_config.maps_to == metatype:
            return True
    return False


def get_fields(cls: type) -> List[str]:
    """
    Get the fields in a class.

    :param cls: The class to get the fields from.
    :returns: fields present in the inputs class, as a list of strings
    """

    fields: list[str] = []
    for attribute in cls.__dict__:
        if not attribute.startswith('_'):
            fields.append(attribute)

    return fields


def infer_descriptor_rows(table_config: TableConfig) -> List[ROW]:
    """
    Infers SchemaSheet descriptor rows, as required by SchemaExporter, for a given TableConfig.

    :param table_config: The TableConfig to infer the descriptor rows for.
    :returns: A list of descriptor rows.
    """

    cs_fields = get_fields(ColumnSettings)
    cs_fields.sort()

    desired_schemasheets_columns = ["header", ] + cs_fields

    descriptor_rows: list[dict[str, str]] = []

    for schemasheet_col in desired_schemasheets_columns:
        index = 0
        temp_dict = {}
        i_s_count = 0
        i_k_count = 0
        for tcck, tccv in table_config.columns.items():
            prefix = ''
            if index == 0:
                prefix = '>'

            # todo differentiate between a verbatim header and a slugged element_name
            if schemasheet_col == "header":
                temp_dict[tcck] = f"{prefix}{tccv.name}"

            elif schemasheet_col == "internal_separator":
                i_s = tccv.settings.internal_separator
                if i_s:
                    temp_dict[tcck] = f'{prefix}internal_separator: "{i_s}"'
                    i_s_count += 1
                else:
                    temp_dict[tcck] = f'{prefix}'

            elif schemasheet_col == "inner_key":
                i_k = tccv.settings.inner_key
                if i_k:
                    temp_dict[tcck] = f'{prefix}inner_key: "{i_k}"'
                    i_k_count += 1
                else:
                    temp_dict[tcck] = f'{prefix}'

            index += 1

        if schemasheet_col == "internal_separator" and i_s_count == 0:
            temp_dict = {}

        if schemasheet_col == "inner_key" and i_k_count == 0:
            temp_dict = {}

        if temp_dict:
            descriptor_rows.append(temp_dict)

    return descriptor_rows


@dataclass
class SchemaExporter:
    """
    Exports a schema to Schema Sheets TSV format
    """
    schemamaker: SchemaMaker = field(default_factory=lambda: SchemaMaker())
    delimiter = '\t'
    rows: List[ROW] = field(default_factory=lambda: [])

    def export(self, schemaview: SchemaView, to_file: Union[str, Path], specification: str = None,
               table_config: TableConfig = None):
        """
        Exports a schema to a schemasheets TSV

        EITHER a specification OR (a table_config and descriptor_rows) must be passed.
        This informs how schema elements are mapped to rows

        :param schemaview:
        :param specification:
        :param to_file:
        :param table_config:
        :return:
        """

        if specification is not None:
            schemasheet = SchemaSheet.from_csv(specification, delimiter=self.delimiter)
            table_config = schemasheet.table_config
            descriptor_rows = schemasheet.table_config_rows
            logging.info(f'Remaining rows={len(schemasheet.rows)}')
        elif table_config is not None:
            descriptor_rows = infer_descriptor_rows(table_config)
        else:
            raise ValueError("Must specify EITHER specification OR table_config")
        for prefix in schemaview.schema.prefixes.values():
            self.export_element(prefix, None, schemaview, table_config)
        for slot in schemaview.all_slots().values():
            self.export_element(slot, None, schemaview, table_config)
        if _configuration_has_primary_keys_for(table_config, T_CLASS):
            for cls in schemaview.all_classes().values():
                self.export_element(cls, None, schemaview, table_config)
                for att in cls.attributes.values():
                    self.export_element(att, cls, schemaview, table_config)
                for su in cls.slot_usage.values():
                    self.export_element(su, cls, schemaview, table_config)
        for e in schemaview.all_enums().values():
            self.export_element(e, None, schemaview, table_config)
            for pv in e.permissible_values.values():
                self.export_element(pv, e, schemaview, table_config)
        for typ in schemaview.all_types().values():
            self.export_element(typ, None, schemaview, table_config)
        for subset in schemaview.all_subsets().values():
            self.export_element(subset, None, schemaview, table_config)

        with open(to_file, 'w', encoding='utf-8') as stream:
            writer = csv.DictWriter(
                stream,
                delimiter=self.delimiter,
                fieldnames=table_config.columns.keys())
            writer.writeheader()

            for row in descriptor_rows:
                writer.writerow(row)

            for row in self.rows:
                writer.writerow(row)

    def export_element(self, element: Element, parent: Optional[Element], schemaview: SchemaView,
                       table_config: TableConfig):
        """
        Translates an individual schema element to a row

        A row is either a simple row representing a standalone element, or it represents a contextualized element, in
        which case a *parent* element is also provided.

        - A PermissibleValue element *MUST* be contextualized using a parent EnumDefinition
        - A SlotDefinition element *MAY* be contextualized using a parent ClassDefinition

        :param element: the element to be exported, e.g an instance of SlotDefinition, ClassDefinition, ...
        :param parent: contextual element; for slots, the parent may be a class; for permissible value, an Enum
        :param schemaview:
        :param table_config:
        :return:
        """

        # Step 1: determine both primary key (pk) column, a pk of any parent
        pk_col = None
        parent_pk_col = None
        for col_name, col_config in table_config.columns.items():
            if col_config.is_element_type:
                t = col_config.maps_to
                if t == T_CLASS:
                    # slots MAY be contextualized by classes
                    if isinstance(element, ClassDefinition):
                        pk_col = col_name
                    if isinstance(parent, ClassDefinition):
                        parent_pk_col = col_name
                elif t == T_SLOT:
                    if isinstance(element, SlotDefinition):
                        pk_col = col_name
                    else:
                        continue
                elif t == T_TYPE:
                    if isinstance(element, TypeDefinition):
                        pk_col = col_name
                    else:
                        continue
                elif t == T_SUBSET:
                    if isinstance(element, SubsetDefinition):
                        pk_col = col_name
                    else:
                        continue
                elif t == T_ENUM:
                    # permissible values MUST be contextualized by enums
                    if isinstance(element, EnumDefinition):
                        pk_col = col_name
                    if isinstance(parent, EnumDefinition):
                        parent_pk_col = col_name
                elif t == T_PV:
                    if isinstance(element, PermissibleValue):
                        pk_col = col_name
                    else:
                        continue
                elif t == T_PREFIX:
                    if isinstance(element, Prefix):
                        pk_col = col_name
                    else:
                        continue
                else:
                    raise AssertionError(f"Unexpected type: {t}")
        if not pk_col:
            logging.info(f"Skipping element: {element}, no PK")
            return
        # Step 2: iterate through all columns in the spec, and populate a row object
        exported_row = {}
        for col_name, col_config in table_config.columns.items():
            settings = col_config.settings
            # Either: (1) this column is mapped to a metamodel slot (metaslot), or
            # (2) the column is a type designator (e.g. holds a value like "class" or "slot")
            if col_config.metaslot:
                # Lookup the value of the element for this metaslot;
                # e.g. if element = SlotDefinition('phone_no', range='string'), then:
                #  - if the column has a metaslot 'name', v='phone no'
                #  - if the column has a metaslot 'range', v='string'
                v = getattr(element, underscore(col_config.metaslot.name), None)
                if v is not None and v != [] and v != {}:
                    # TODO: consider moving this to a standalone function
                    # inner function to map an atomic value
                    def repl(v: Any) -> Optional[str]:
                        if col_config.maps_to == 'examples':
                            if isinstance(v, Example):
                                return v.value
                            else:
                                raise ValueError(f"Expected Example, got {type(v)} for {v}")
                        if col_config.settings.inner_key:
                            if isinstance(v, Annotation):
                                if v.tag == col_config.settings.inner_key:
                                    return v.value
                                else:
                                    return None
                            else:
                                v = getattr(v, col_config.settings.inner_key, None)
                                if isinstance(v, bool):
                                    v = str(v).lower()
                                return v
                        if settings.curie_prefix:
                            pfx = f'{settings.curie_prefix}:'
                            if v.startswith(pfx):
                                return v.replace(pfx, '')
                            else:
                                return None
                        if isinstance(v, bool):
                            return str(v).lower()
                        return v

                    # map the value (which may be a collection or an object) to a flat string
                    # representation
                    if isinstance(v, list):
                        v = [repl(v1) for v1 in v if repl(v1) is not None]
                        v = '|'.join([str(i) for i in v])
                        if v != '':
                            exported_row[col_name] = v
                    elif isinstance(v, dict):
                        v = [repl(v1) for v1 in v.values() if repl(v1) is not None]
                        v = '|'.join([str(i) for i in v])
                        if v != '':
                            exported_row[col_name] = v
                    else:
                        v = repl(v)
                        if v is not None:
                            exported_row[col_name] = str(v)
            elif col_config.is_element_type:
                # the column holds the metatype of the element;
                # e.g if slot=SlotDefinition(...), then the value of a column
                # 'type' that is a type designator, then the value will be 'slot'
                if pk_col == col_name:
                    if isinstance(element, PermissibleValue):
                        # permissible values are treated differently from other metamodel
                        # elements, as they have no name
                        exported_row[col_name] = element.text
                        if not parent_pk_col:
                            raise ValueError(f"Cannot have floating permissible value {element.text}")
                    elif isinstance(element, Prefix):
                        exported_row[col_name] = element.prefix_prefix
                    else:
                        exported_row[col_name] = element.name
                elif parent_pk_col == col_name:
                    exported_row[col_name] = parent.name
                else:
                    logging.info(f'TODO: {col_name} [{type(element).class_name}] // {col_config}')
            else:
                logging.info(f'IGNORING: {col_name} // {col_config}')
        self.export_row(exported_row)

    def export_row(self, row: ROW):
        self.rows.append(row)

    def is_slot_redundant(self, slot: SlotDefinition, schemaview: SchemaView):
        for c in schemaview.all_classes().values():
            if slot.name in c.slots:
                pass


@click.command()
@click.option('-o', '--output',
              help="output file")
@click.option("-d", "--output-directory",
              help="folder in which to store resulting TSVs")
@click.option("-s", "--schema",
              required=True,
              help="Path to the schema")
@click.option("--overwrite/--no-overwrite",
              default=False,
              show_default=True,
              help="If set, then overwrite existing schemasheet files if they exist")
@click.option("--append-sheet/--no-append-sheet",
              default=False,
              show_default=True,
              help="If set, then append to existing schemasheet files if they exist")
@click.option("--unique-slots/--no-unique-slots",
              default=False,
              show_default=True,
              help="All slots are treated as unique and top level and do not belong to the specified class")
@click.option("-v", "--verbose", count=True)
@click.argument('tsv_files', nargs=-1)
def export_schema(tsv_files, output_directory, output: TextIO, overwrite: bool, append_sheet: bool,
                  schema, unique_slots: bool, verbose: int):
    """
    Convert LinkML schema to schemasheets

    Convert a schema to a single sheet, writing on stdout:

        linkml2sheets -s my_schema.yaml my_schema_spec.tsv > my_schema.tsv

    As above, with explicit output:

        linkml2sheets -s my_schema.yaml my_schema_spec.tsv -o my_schema.tsv

    Convert schema to multisheets, writing output to a folder:

        linkml2sheets -s my_schema.yaml specs/*.tsv -d output

    Convert schema to multisheets, writing output in place:

        linkml2sheets -s my_schema.yaml sheets/*.tsv -d sheets --overwrite

    Convert schema to multisheets, appending output:

        linkml2sheets -s my_schema.yaml sheets/*.tsv -d sheets --append


    """
    if verbose >= 2:
        logging.basicConfig(level=logging.DEBUG)
    elif verbose == 1:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)
    if output is not None and output_directory:
        raise ValueError(f'Cannot combine output-directory and output options')
    if output is not None and len(tsv_files) > 1:
        raise ValueError(f'Cannot use output option with multiple sheets')
    if append_sheet:
        raise NotImplementedError(f'--append-sheet not yet implemented')
    exporter = SchemaExporter()
    sv = SchemaView(schema)
    for f in tsv_files:
        if output_directory:
            outpath: Path = Path(output_directory) / Path(f).name
        else:
            if output is not None:
                outpath = Path(output)
            else:
                outpath = sys.stdout
        if isinstance(outpath, Path) and outpath.exists():
            if overwrite:
                logging.info(f'Overwriting: {outpath}')
            else:
                raise PermissionError(f'Will not overwrite {outpath} unless --overwrite is set')
        exporter.export(sv, specification=f, to_file=outpath)


if __name__ == '__main__':
    export_schema()



from typing import Optional

from bioregistry import get_iri

priority = ["obofoundry", "default", "miriam", "ols", "n2t", "bioportal"]


def guess_prefix_expansion(prefix: str) -> Optional[str]:
    """
    Guesses a prefix expansion using bioregistry

    :param prefix:
    :return:
    """
    try:
        return get_iri(prefix, "", priority=priority)
    except KeyError:
        return None


BASE = "https://docs.google.com/spreadsheets/d"

def gsheets_download_url(sheet_id: str, sheet_name: str) -> str:
    """
    Get the google sheets download URL

    :param sheet_id: 1wVoaiFg47aT9YWNeRfTZ8tYHN8s8PAuDx5i2HUcDpvQ
    :param sheet_name: e.g. personinfo
    :return:
    """
    return f"{BASE}/{sheet_id}/gviz/tq?tqx=out:csv&sheet={sheet_name}"





import csv
from dataclasses import dataclass
from typing import Union, Dict, List, Any
import pkgutil
from pathlib import PurePath, Path
from functools import lru_cache
import logging
import yaml
from linkml_runtime.linkml_model import SlotDefinition, ClassDefinition, SchemaDefinition, \
    PermissibleValue, EnumDefinition, TypeDefinition, SubsetDefinition, Prefix
from linkml_runtime.linkml_model.meta import Setting

from linkml_runtime.utils.schemaview import SchemaView

from schemasheets.conf.configschema import ColumnSettings, Shortcuts

COL_NAME = str
DESCRIPTOR = str
ROW = Dict[str, Any]


#c = ClassDefinition
T_SCHEMA = 'schema'
T_CLASS = 'class'
T_SLOT = 'slot'
T_ENUM = 'enum'
T_PV = 'permissible_value'
T_TYPE = 'type'
T_SUBSET = 'subset'
T_PREFIX = 'prefix'
T_SETTING = 'setting'

tmap = {
    T_SCHEMA: SchemaDefinition,
    T_CLASS: ClassDefinition,
    T_SLOT: SlotDefinition,
    T_ENUM: EnumDefinition,
    T_PV: PermissibleValue,
    T_TYPE: TypeDefinition,
    T_SUBSET: SubsetDefinition,
    T_PREFIX: Prefix,
    T_SETTING: Setting,
}




@dataclass
class ColumnConfig:
    """
    Configuration for a single column in a schema sheet
    """
    name: COL_NAME
    maps_to: DESCRIPTOR = None
    settings: ColumnSettings = None
    metaslot: SlotDefinition = None
    inner_key_metaslot: SlotDefinition = None
    is_element_type: bool = None

    def merge_settings(self, settings: ColumnSettings) -> None:
        """
        merges specified settings into current settings

        :param settings: settings to be merged
        """
        for k, v in vars(settings).items():
            if v:
                setattr(self.settings, k, v)

    def add_info(self, info: Union[Dict, DESCRIPTOR]) -> None:
        """
        Adds configuration/settings in the form of a dict object.

        Information can be incrementally added:

        - the first piece of information should be the descriptor
        - after that individual settings can be added

        :param info: configuration
        :return:
        """
        logging.debug(f"Adding info to {self.name}: {info}")
        if self.maps_to is None:
            # The first descriptor row describes what the column maps to
            self.settings = ColumnSettings()
            if isinstance(info, dict):
                items = list(info.items())
                if len(items) != 1:
                    raise ValueError(f'Unexpected settings: {info}')
                else:
                    item = items[0]
                    self.maps_to = item[0]
                    if isinstance(item[1], dict):
                        settings = ColumnSettings(**item[1])
                    else:
                        raise ValueError(f'Expected dict after first element of {items}')
                    self.merge_settings(settings)
            else:
                self.maps_to = info
            mm = get_metamodel()
            snmap = mm.slot_name_mappings()
            # TODO: use alias
            snmap['uri'] = snmap['type_uri']
            if self.maps_to.startswith("metaslot."):
                maps_to = self.maps_to.replace("metaslot.", "")
                self.metaslot = snmap[maps_to]
                self.maps_to = maps_to
            elif self.maps_to in snmap and self.maps_to != 'type':
                self.metaslot = snmap[self.maps_to]
            else:
                if self.maps_to not in tmap and self.maps_to not in Shortcuts:
                    raise ValueError(f'Cannot interpret: {self.maps_to}')
        else:
            settings = ColumnSettings(**info)
            self.merge_settings(settings)
            if settings.inner_key:
                snmap = get_metamodel().slot_name_mappings()
                if settings.inner_key in snmap:
                    self.inner_key_metaslot = snmap[settings.inner_key]



@dataclass
class TableConfig:
    """
    Configuration for an entire table / schema sheet

    """
    name: str = None
    """table name"""

    columns: Dict[COL_NAME, ColumnConfig] = None
    """maps column names to config"""

    column_by_element_type: Dict[str, COL_NAME] = None
    """maps element types (schema, class, ...) to the name of the column that represents them"""

    metatype_column: COL_NAME = None
    """Column that represents the metatype designator"""

    name_column: COL_NAME = None
    """Column that represents that name of the entity"""

    def add_info(self, col: COL_NAME, info: Union[Dict, DESCRIPTOR]) -> None:
        """
        Wrapper for :ref:`ColumnConfig.add_info`

        :param col:
        :param info:
        """
        if col not in self.columns:
            self.columns[col] = ColumnConfig(col)
        #print(f'ADDING: {col}')
        self.columns[col].add_info(info)
        if self.columns[col].maps_to == 'metatype':
            if self.metatype_column and self.metatype_column != col:
                raise ValueError(f'Multiple metatype columns not allowed: {self.metatype_column}, {col}')
            self.metatype_column = col
        if self.columns[col].maps_to == 'name':
            if self.name_column:
                raise ValueError(f'Multiple name columns not allowed: {self.name_column}, {col}')
            self.name_column = col
        if self.column_by_element_type is None:
            self.column_by_element_type = {}
        for c, cc in self.columns.items():
            if cc.maps_to in tmap:
                self.column_by_element_type[cc.maps_to] = c
                cc.is_element_type = True


@dataclass
class SchemaSheet:
    """
    A SchemaSheet consists of:

    - a collection of rows, each row representing a schema element
    - a TableConfiguration
    """
    table_config: TableConfig
    rows: List[ROW]
    start_line_number: int
    table_config_rows: List[ROW] = None

    @classmethod
    def from_csv(cls, path: str, delimiter='\t'):
        with open(path, newline='') as tsv_file:
            reader = csv.DictReader(tsv_file, delimiter=delimiter)
            return cls.from_dictreader(reader)

    @staticmethod
    def from_dictreader(reader: csv.DictReader) -> "SchemaSheet":
        """
        Reads a schemasheets TSV file parsing only header info

        :param reader:
        :return:
        """
        table_config = TableConfig(columns={})
        rows = []
        line_num = 1
        table_config_rows = []
        for row in reader:
            logging.debug(f"ROW: {row}")
            # google sheets
            if "" in row:
                del row[""]
            k0 = list(row.keys())[0]
            if row[k0].startswith('>'):
                table_config_rows.append(row)
                line_num += 1
                for k, v in row.items():
                    if v is not None and v.startswith('>'):
                        v = v.replace('>', '')
                    if v:
                        meta_obj = yaml.safe_load(v)
                        table_config.add_info(k, meta_obj)
                    else:
                        if line_num == 2:
                            # TODO: consider auto-interpreting
                            raise ValueError(f'Enter an interpretation for {k}')
                        logging.debug(f'Empty val for {k} in line {line_num}')
            else:
                rows.append(row)
        return SchemaSheet(table_config=table_config,
                           table_config_rows=table_config_rows,
                           rows=rows,
                           start_line_number=line_num)

    def load_table_config(self, config: Union[dict, str, Path]) -> None:
        """
        Loads a table configuration from a file or dict

        :param config:
        :return:
        """
        if not isinstance(config, dict):
            with open(config) as f:
                config = yaml.safe_load(f)
                return self.load_table_config(config)
        for k, v in config.items():
            if isinstance(v, list):
                for v1 in v:
                    self.table_config.add_info(k, v1)
            else:
                self.table_config.add_info(k, v)

@lru_cache()
def get_metamodel() -> SchemaView:
    """
    Returns the LinkML schema metamodel as a SchemaView object

    this can be retired when https://github.com/linkml/linkml-runtime/pull/100/
    is in major release
    :return:
    """
    package = 'linkml_runtime.linkml_model.meta'
    full_path = PurePath('model') / 'schema'
    data = pkgutil.get_data(package, f'{full_path}/meta.yaml')
    return SchemaView(data.decode("utf-8"))

@lru_cache()
def get_configmodel() -> SchemaView:
    """
    Returns the Config schema metamodel as a SchemaView object

    this can be retired when https://github.com/linkml/linkml-runtime/pull/100/
    is in major release
    :return:
    """
    package = 'schemasheets.conf.configschema'
    data = pkgutil.get_data(package, f'configschema.yaml')
    return SchemaView(data.decode("utf-8"))


import codecs
import contextlib
import os
import sys
import csv
import logging
import tempfile
from urllib.request import urlopen

import click
import yaml
from dataclasses import dataclass
from typing import List, Union, Any, Dict, Tuple, Generator, TextIO

from linkml_runtime.dumpers import yaml_dumper
from linkml_runtime.linkml_model import Annotation, Example
from linkml_runtime.linkml_model.meta import SchemaDefinition, ClassDefinition, Prefix, \
    SlotDefinition, EnumDefinition, PermissibleValue, SubsetDefinition, TypeDefinition, Element, Setting
from linkml_runtime.utils.schema_as_dict import schema_as_dict
from linkml_runtime.utils.schemaview import SchemaView, re

from schemasheets.schemasheet_datamodel import ColumnConfig, TableConfig, get_configmodel, get_metamodel, COL_NAME, \
    DESCRIPTOR, \
    tmap, T_CLASS, T_PV, T_SLOT, T_SUBSET, T_SCHEMA, T_ENUM, T_PREFIX, T_TYPE, SchemaSheet, T_SETTING
from schemasheets.conf.configschema import Cardinality
from schemasheets.utils.google_sheets import gsheets_download_url
from schemasheets.utils.prefixtool import guess_prefix_expansion


class SchemaSheetRowException(Exception):
    pass


@dataclass
class SchemaMaker:
    """
    Engine for making LinkML schemas from Schema Sheets
    """
    schema: SchemaDefinition = None
    element_map: Dict[Tuple[str, str], Element] = None
    metamodel: SchemaView = None
    cardinality_vocabulary: str = None
    use_attributes: bool = None
    default_name: str = None
    unique_slots: bool = None
    gsheet_id: str = None
    table_config_path: str = None

    def create_schema(self, csv_files: Union[str, List[str]], **kwargs) -> SchemaDefinition:
        """
        Create a LinkML schema from a collection of Schema Sheets

        :param csv_files: schema sheets
        :param kwargs:
        :return: generated schema
        """
        n = self.default_name
        if n is None:
            n = 'TEMP'
        self.schema = SchemaDefinition(id=n, name=n, default_prefix=n, default_range='string')
        if not isinstance(csv_files, list):
            csv_files = [csv_files]
        for f in csv_files:
            self.merge_sheet(f, **kwargs)
        self.schema.imports.append('linkml:types')
        self.schema.prefixes['linkml'] = Prefix('linkml', 'https://w3id.org/linkml/')
        self._tidy_slot_usage()
        prefix = self.schema.default_prefix
        if prefix not in self.schema.prefixes:
            logging.error(f'Prefix {prefix} not declared: using default')
            self.schema.prefixes[prefix] = Prefix(prefix, f'https://example.org/{prefix}/')
        return self.schema

    def _tidy_slot_usage(self):
        """
        removes all slot usages marked inapplicable

        :return:
        """
        for cn, c in self.schema.classes.items():
            inapplicable_slots = [sn for sn, s in c.slot_usage.items() if 'inapplicable' in s.annotations]
            for sn in inapplicable_slots:
                c.slots.remove(sn)
                del c.slot_usage[sn]

    def merge_sheet(self, file_name: str, delimiter='\t') -> None:
        """
        Merge information from the given schema sheet into the current schema

        :param file_name: schema sheet
        :param delimiter: default is tab
        :return:
        """
        logging.info(f'READING {file_name} D={delimiter}')
        # with self.ensure_file(file_name) as tsv_file:
        #    reader = csv.DictReader(tsv_file, delimiter=delimiter)
        with self.ensure_csvreader(file_name, delimiter=delimiter) as reader:
            schemasheet = SchemaSheet.from_dictreader(reader)
            if self.table_config_path:
                schemasheet.load_table_config(self.table_config_path)
            line_num = schemasheet.start_line_number
            # TODO: check why this doesn't work
            # while rows and all(x for x in rows[-1] if not x):
            #    print(f'TRIMMING: {rows[-1]}')
            #    rows.pop()
            logging.info(f'ROWS={len(schemasheet.rows)}')
            for row in schemasheet.rows:
                try:
                    self.add_row(row, schemasheet.table_config)
                    line_num += 1
                except ValueError as e:
                    raise SchemaSheetRowException(f'Error in line {line_num}, row={row}') from e

    def add_row(self, row: Dict[str, Any], table_config: TableConfig):
        for element in self.row_focal_element(row, table_config):
            if isinstance(element, Prefix):
                name = element.prefix_prefix
            elif isinstance(element, PermissibleValue):
                name = element.text
            elif isinstance(element, Setting):
                # print(f"\n{element = }")
                name = element.setting_key
            else:
                logging.debug(f'EL={element} in {row}')
                name = element.name
            logging.debug(f'ADDING: {row} // {name}')
            for k, v in row.items():
                # print(f"\n{k = }")
                if k not in table_config.columns:
                    raise ValueError(f'Expected to find {k} in {table_config.columns.keys()}')
                cc = table_config.columns[k]
                # print(f"{cc = }")
                v = self.normalize_value(v, cc)
                if v:
                    # print(f"{v = }")
                    # special case: class-context provided by settings
                    if cc.settings.applies_to_class:
                        actual_element = list(self.row_focal_element(row, table_config, column=k))[0]
                    else:
                        actual_element = element
                    # print(f"{cc.maps_to = }")
                    # print(f"{cc = }")
                    logging.debug(f'SETTING {name} {cc.maps_to} = {v}')
                    if cc.maps_to == 'cardinality':
                        self.set_cardinality(actual_element, v)
                    elif cc.metaslot:
                        if cc.maps_to == 'examples':
                            for vi in v:
                                actual_element.examples.append(Example(value=vi))
                        elif cc.maps_to == 'annotations':
                            if cc.settings.inner_key:
                                ann = Annotation(cc.settings.inner_key, v)
                                actual_element.annotations[ann.tag] = ann
                            else:
                                anns = yaml.safe_load(v[0])
                                for ann_key, ann_val in anns.items():
                                    actual_element.annotations[ann_key] = ann_val
                        elif isinstance(v, list):
                            setattr(actual_element, cc.maps_to, getattr(actual_element, cc.maps_to, []) + v)
                        elif isinstance(v, dict):
                            for v_k, v_v in v.items():
                                curr_dict = getattr(actual_element, cc.maps_to)
                                curr_dict[v_k] = v_v
                        else:
                            if cc.settings.inner_key:
                                # inner keys settings allow for a flat value to map to an internal part of an
                                # element, such as an annotation value or structured value syntax
                                curr_obj = getattr(actual_element, cc.maps_to)
                                if curr_obj is None:
                                    # for some slots such as structured_pattern, there is no default
                                    # object set; create an empty one here as dict.
                                    # will later be converted to a metamodel object
                                    curr_obj = {}
                                    setattr(actual_element, cc.maps_to, curr_obj)
                                if isinstance(curr_obj, dict):
                                    curr_val = curr_obj.get(cc.settings.inner_key, None)
                                else:
                                    # https://github.com/linkml/linkml/issues/971
                                    curr_val = getattr(curr_obj, cc.settings.inner_key, None)
                            else:
                                curr_val = getattr(actual_element, cc.maps_to)
                            # print(f"{curr_val = }")
                            # print(f"{v = }")

                            if curr_val and curr_val != 'TEMP' and curr_val != v and \
                                    not isinstance(actual_element, SchemaDefinition) and \
                                    not isinstance(actual_element, Prefix) and \
                                    not isinstance(actual_element, Setting):
                                logging.warning(f'Overwriting value for {k}, was {curr_val}, now {v}')
                                raise ValueError(f'Cannot reset value for {k}, was {curr_val}, now {v}')
                            if cc.settings.inner_key:
                                getattr(actual_element, cc.maps_to)[cc.settings.inner_key] = v
                            else:
                                setattr(actual_element, cc.maps_to, v)
                    elif cc.is_element_type:
                        logging.debug(f'Already accounted for {k}')
                    elif cc.maps_to == 'metatype':
                        logging.debug(f'Already accounted for {k}')
                    elif cc.maps_to == 'ignore':
                        logging.debug(f'Ignoring {k}')
                    else:
                        raise ValueError(f'No mapping for {k}; cc={cc}')

    def get_current_element(self, elt: Element) -> Union[Element, PermissibleValue]:
        """
        Look up an element in the current schema using a stub element as key

        If an element cannot be found, then the stub is added to the schema

        For example, the first time this is called:

        > el = sm.get_current_element(ClassDefinition("foo"))

        A class "foo" will be added to the schema, and the input will be returned as
        the output.

        This object may then later be adorned with other elements coming from a schemasheet row;
        description, slots, ...

        subsequently, the method may be called again:

        > el = sm.get_current_element(ClassDefinition("foo"))

        This time the existing "foo" class from the schema, with its adornments, will be returned

        :param elt: proxy for element to look up
        :return:
        """
        sc = self.schema
        if isinstance(elt, SchemaDefinition):
            # TODO: consider multiple schemas per sheet
            return sc
        elif isinstance(elt, PermissibleValue):
            return elt
        else:
            if isinstance(elt, ClassDefinition):
                ix = sc.classes
            elif isinstance(elt, SlotDefinition):
                ix = sc.slots
            elif isinstance(elt, EnumDefinition):
                ix = sc.enums
            elif isinstance(elt, TypeDefinition):
                ix = sc.types
            elif isinstance(elt, SubsetDefinition):
                ix = sc.subsets
            else:
                raise ValueError(f'TODO: implement for type {type(elt)} in {elt}')
            if elt.name not in ix:
                logging.debug(f"Adding {elt.name} to schema")
                ix[elt.name] = elt
            return ix[elt.name]

    def row_focal_element(self, row: Dict[str, Any], table_config: TableConfig,
                          column: COL_NAME = None) -> Generator[None, Element, None]:
        """
        For a given row, yield one or more metamodel elements

        Typically a row will have a single focal element; i.e. the row represents a class, or
        a slot, or a type, or a prefix...

        But rows can also double-up

        :param row:
        :param table_config:
        :return:
        """
        vmap: Dict[str, List[Element]] = {}
        main_elt = None
        if table_config.metatype_column:
            tc = table_config.metatype_column
            if tc in row:
                typ = self.normalize_value(row[tc], table_config.columns[tc])
                if not table_config.name_column:
                    raise ValueError(f'name column must be set when type column ({tc}) is set; row={row}')
                name_val = row[table_config.name_column]
                if not name_val:
                    raise ValueError(f'name column must be set when type column ({tc}) is set')
                if typ == 'class':
                    vmap[T_CLASS] = [self.get_current_element(ClassDefinition(name_val))]
                elif typ == 'slot':
                    vmap[T_SLOT] = [self.get_current_element(SlotDefinition(name_val))]
                else:
                    raise ValueError(f'Unknown metatype: {typ}')
        if table_config.column_by_element_type is None:
            raise ValueError(f'No table_config.column_by_element_type')
        for k, elt_cls in tmap.items():
            if k in table_config.column_by_element_type:
                col = table_config.column_by_element_type[k]
                if col in row:
                    v = self.normalize_value(row[col])
                    if v:
                        if '|' in v:
                            vs = v.split('|')
                        else:
                            vs = [v]
                        if elt_cls == Prefix:
                            if len(vs) != 1:
                                raise ValueError(f'Cardinality of prefix col must be 1; got: {vs}')
                            pfx = Prefix(vs[0], 'TODO')
                            self.schema.prefixes[pfx.prefix_prefix] = pfx
                            vmap[k] = [pfx]
                        elif elt_cls == Setting:
                            if len(vs) != 1:
                                raise ValueError(f'Cardinality of setting col must be 1; got: {vs}')
                            stg = Setting(vs[0], 'TODO')
                            self.schema.settings[stg.setting_key] = stg
                            vmap[k] = [stg]
                        elif elt_cls == SchemaDefinition:
                            if len(vs) != 1:
                                raise ValueError(f'Cardinality of schema col must be 1; got: {vs}')
                            self.schema.name = vs[0]
                            vmap[k] = [self.schema]
                        else:
                            vmap[k] = [self.get_current_element(elt_cls(v)) for v in vs]

        def check_excess(descriptors):
            diff = set(vmap.keys()) - set(descriptors + [T_SCHEMA])
            if len(diff) > 0:
                raise ValueError(f'Excess slots: {diff}')

        if column:
            cc = table_config.columns[column]
            if cc.settings.applies_to_class:
                if T_CLASS in vmap and vmap[T_CLASS]:
                    raise ValueError(f'Cannot use applies_to_class in class-focused row')
                else:
                    cls = self.get_current_element(ClassDefinition(cc.settings.applies_to_class))
                    vmap[T_CLASS] = [cls]
        if T_SLOT in vmap:
            check_excess([T_SLOT, T_CLASS])
            if len(vmap[T_SLOT]) != 1:
                raise ValueError(f'Cardinality of slot field must be 1; got {vmap[T_SLOT]}')
            main_elt = vmap[T_SLOT][0]
            if T_CLASS in vmap:
                # The sheet does double duty representing a class and a slot;
                # Here *both* the "class" and "slot" columns are populated, so
                # this translated to slot_usage;
                # TODO: add option to allow to instead represent these as attributes
                c: ClassDefinition
                for c in vmap[T_CLASS]:
                    if self.use_attributes:
                        # slots always belong to a class;
                        # no separate top level slots
                        a = SlotDefinition(main_elt.name)
                        c.attributes[main_elt.name] = a
                        yield a
                    else:
                        # add top level slot if not present
                        if main_elt.name not in c.slots:
                            c.slots.append(main_elt.name)
                        if self.unique_slots:
                            yield main_elt
                        else:
                            c.slot_usage[main_elt.name] = SlotDefinition(main_elt.name)
                            main_elt = c.slot_usage[main_elt.name]
                            yield main_elt
            else:
                yield main_elt
        elif T_CLASS in vmap:
            # This row represents a class element
            # (note if the sheet does double duty for classes and slots, this particular
            #  row is *only* about the slot)
            check_excess([T_CLASS])
            for main_elt in vmap[T_CLASS]:
                yield main_elt
        elif T_ENUM in vmap:
            check_excess([T_ENUM, T_PV])
            if len(vmap[T_ENUM]) != 1:
                raise ValueError(f'Cardinality of enum field must be 1; got {vmap[T_ENUM]}')
            this_enum: EnumDefinition = vmap[T_ENUM][0]
            if T_PV in vmap:
                for pv in vmap[T_PV]:
                    # pv = PermissibleValue(text=v)
                    this_enum.permissible_values[pv.text] = pv
                    yield pv
            else:
                yield this_enum
        elif T_PREFIX in vmap:
            for main_elt in vmap[T_PREFIX]:
                yield main_elt
        elif T_TYPE in vmap:
            for main_elt in vmap[T_TYPE]:
                yield main_elt
        elif T_SUBSET in vmap:
            for main_elt in vmap[T_SUBSET]:
                yield main_elt
        elif T_SCHEMA in vmap:
            for main_elt in vmap[T_SCHEMA]:
                yield main_elt
        elif T_SETTING in vmap:
            for main_elt in vmap[T_SETTING]:
                yield main_elt
        else:
            raise ValueError(f'Could not find a focal element for {row}')

    def normalize_value(self, v: str, column_config: ColumnConfig = None) -> Any:
        """
        Given a row/column value, normalize it according to general rules and sheet settings

        General rules:

        - The strings "", ".", and "n/a" will be treated as if empty/None
        - for boolean descriptors, the values "yes" and "no" are permissible for True/False
        - leading and trailing whitespace is trimmed/stripped

        Specific settings:

        - See configschema for all options

        For example, if this method is called with a value v and a column config that has
        a regex pattern, then the regex is used to extract the value from v

        :param v:
        :param column_config: optional
        :return:
        """
        if column_config:
            metaslot = column_config.metaslot
        else:
            metaslot = None
        if v == '.':
            v = None
        elif v == '':
            v = None
        elif v == 'n/a':
            v = None
        if v and (v.startswith(' ') or v.endswith(' ')):
            if column_config:
                logging.warning(f'Stripping value: "{v}" for {column_config.name}')
            else:
                logging.warning(f'Stripping value: "{v}" (no column config)')
            v = v.strip()
        if v and column_config:
            re_match = column_config.settings.regular_expression_match
            if re_match:
                m = re.search(re_match, v)
                if m:
                    v = m.group(1)
                else:
                    logging.error(f'No match using {column_config.settings.regular_expression_match} on {v}')
                    v = None
            if column_config.settings.curie_prefix:
                if ':' in v:
                    logging.warning(
                        f'Will not prefix {v} with {column_config.settings.curie_prefix} as it is already prefixed')
                else:
                    v = f'{column_config.settings.curie_prefix}:{v}'
            if column_config.settings.prefix:
                v = f'{column_config.settings.prefix}{v}'
            if column_config.settings.suffix:
                v = f'{column_config.settings.suffix}{v}'
            if column_config.settings.vmap:
                vmap = column_config.settings.vmap
                if v in vmap:
                    v = vmap[v].map_value
                elif '*' in vmap:
                    v = vmap['*'].map_value
                else:
                    logging.warning(f'No mapping for {v}, passing through')
        if metaslot and metaslot.range:
            rng = metaslot.range
            if column_config.inner_key_metaslot:
                rng = column_config.inner_key_metaslot.range
            if rng == 'boolean':
                bmap = {
                    'yes': True,
                    'no': False,
                    'true': True,
                    'false': False,
                }
                if v and v.lower() in bmap:
                    v = bmap[v.lower()]
                else:
                    v = bool(v)
        if metaslot and metaslot.multivalued and not column_config.settings.inner_key:
            if not isinstance(v, list):
                if v is None:
                    v = []
                else:
                    if column_config.settings.internal_separator:
                        v = v.split(column_config.settings.internal_separator)
                    else:
                        v = [v]
        return v

    def set_cardinality(self, element: SlotDefinition, card: str) -> None:
        """
        Sets the cardinality of a slot, allowing a variety of vocabularies.

        See configschema.yaml for all possible vocabularies, these include:

        - UML strings, e.g. '0..1'
        - text strings matching the cardinality vocabulary, e.g. 'zero to one'
        - codes used in cardinality vocabulary, e.g. O, M, ...

        The vocabulary maps to underlying LinkML primitives:

        - required
        - multivalued
        - recommended

        :param element: slot
        :param card: cardinality term
        :return: None
        """
        if not card:
            return
        # shortcuts for UML
        if card == '1':
            card = '1..1'
        if card == '*':
            card = '0..*'
        # parse UML strings as a simple grammar
        if '..' in card:
            [min, max] = card.split('..')
            element.required = int(min) > 0
            element.multivalued = max == '*' or int(max) > 1
        else:
            # lookup in cardinality vocabulary
            sv = get_configmodel()
            cvocab = self.cardinality_vocabulary
            if cvocab is None:
                cvocab = 'code'
            matches = []
            pvs = sv.get_enum('Cardinality').permissible_values
            for pv in pvs.values():
                annv = pv.annotations.get(cvocab)
                if card and annv and annv.value == card:
                    matches.append(pv)
                if card == pv.text:
                    matches.append(pv)
            if len(matches) > 1:
                raise ValueError(f'Ambiguous matches to {card}: {matches}')
            if matches:
                match = matches[0]
                if match.text == Cardinality.not_applicable.text:
                    # this slot usage will be removed post-processing
                    element.annotations['inapplicable'] = Annotation('inapplicable', 'true')
                maps_to_ann = match.annotations.get('maps_to')
                if maps_to_ann:
                    yobj = yaml.safe_load(maps_to_ann.value)
                    for k, v in yobj.items():
                        if k == 'required':
                            element.required = v
                        elif k == 'multivalued':
                            element.multivalued = v
                        elif k == 'recommended':
                            element.recommended = v
            else:
                raise ValueError(f'Cannot parse cardinality: {card} // {pvs.keys()}')

    def repair_schema(self, schema: SchemaDefinition) -> SchemaDefinition:
        """
        Performs repair on schema in place

        - adds default prefixes
        - repairs subsets

        :param schema:
        :return:
        """
        sv = SchemaView(schema)
        # pfx = schema.default_prefix
        # if pfx not in schema.prefixes:
        #    schema.prefixes[pfx] = Prefix(pfx, f'http://example.org/{pfx}/')
        #    logging.info(f'Set default prefix: {schema.prefixes[pfx]}')
        prefixes = set()
        for e in list(sv.all_elements(imports=False).values()):
            # TODO: this does not include slot_usage
            map_ix = sv.get_mappings(e.name)
            for t, curies in map_ix.items():
                if curies:
                    for curie in curies:
                        if ':' in curie:
                            prefixes.add(curie.split(':')[0])
        namespaces = sv.namespaces()
        for pfx in prefixes:
            if pfx not in namespaces:
                logging.debug(f'Guessing prefix expansion: {pfx}')
                pfx_ref = guess_prefix_expansion(pfx)
                if not pfx_ref:
                    pfx_ref = f'http://example.org/{pfx}/'
                schema.prefixes[pfx] = Prefix(pfx, pfx_ref)
                logging.warning(f'Filling in missing prefix for: {pfx} => {pfx_ref}')
        # reset indexes
        sv = SchemaView(schema)
        subsets = set()
        for e in sv.all_elements().values():
            subsets.update(set(e.in_subset))
        for s in subsets:
            if s not in schema.subsets:
                schema.subsets[s] = SubsetDefinition(s)
        return schema

    @contextlib.contextmanager
    def ensure_file(self, file_name: str) -> str:
        if self.gsheet_id:
            url = gsheets_download_url(self.gsheet_id, file_name)
            stream = urlopen(url)
            yield codecs.iterdecode(stream, 'utf-8')
        else:
            with open(file_name) as file:
                yield file

    @contextlib.contextmanager
    def ensure_csvreader(self, file_name: str, delimiter=None) -> str:
        if self.gsheet_id:
            url = gsheets_download_url(self.gsheet_id, file_name)
            stream = urlopen(url)
            text_stream = codecs.iterdecode(stream, 'utf-8')
            reader = csv.DictReader(text_stream, delimiter=",")
            yield reader
        else:
            with open(file_name) as file:
                reader = csv.DictReader(file, delimiter=delimiter)
                yield reader


@click.command()
@click.option('-o', '--output',
              type=click.File(mode="w"),
              default=sys.stdout,
              help="output file")
@click.option('-s', '--sort-keys',
              default=False,
              help="Sort keys in schema output? For example permissible values in an enumeration? Defaults to False.")
@click.option("-n", "--name",
              help="name of the schema")
@click.option("-C", "--table-config-path",
              help="YAML file with header mappings")
@click.option("--unique-slots/--no-unique-slots",
              default=False,
              show_default=True,
              help="All slots are treated as unique and top level and do not belong to the specified class")
@click.option("--use-attributes/--no-use-attributes",
              "-A", "--no-A",
              default=False,
              show_default=True,
              help="All slots specified in conjunction with a class are attributes of that class")
@click.option("--repair/--no-repair",
              default=True,
              show_default=True,
              help="Auto-repair schema")
@click.option("--gsheet-id",
              help="Google sheets ID. If this is specified then the arguments MUST be sheet names")
@click.option("-v", "--verbose", count=True)
@click.argument('tsv_files', nargs=-1)
def convert(tsv_files, gsheet_id, output: TextIO, name, repair, table_config_path: str, use_attributes: bool,
            unique_slots: bool, verbose: int, sort_keys: bool):
    """
    Convert schemasheets to a LinkML schema

    Example:

       sheets2linkml my_schema/*tsv --output my_schema.yaml

    If your sheets are stored as google sheets, then you can pass in --gsheet-id to give the base sheet.
    In this case arguments should be the names of individual tabs

    Example:

        sheets2linkml --gsheet-id 1wVoaiFg47aT9YWNeRfTZ8tYHN8s8PAuDx5i2HUcDpvQ personinfo types -o my_schema.yaml
    """
    if verbose >= 2:
        logging.basicConfig(level=logging.DEBUG)
    elif verbose == 1:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)
    sm = SchemaMaker(use_attributes=use_attributes,
                     unique_slots=unique_slots,
                     gsheet_id=gsheet_id,
                     default_name=name,
                     table_config_path=table_config_path)
    schema = sm.create_schema(list(tsv_files))
    if repair:
        schema = sm.repair_schema(schema)
    schema_dict = schema_as_dict(schema)
    output.write(yaml.dump(schema_dict, sort_keys=sort_keys))
    # output.write(yaml_dumper.dumps(schema))


if __name__ == '__main__':
    convert()
